"""Main script run by server application"""

import time
import traceback
from collections import OrderedDict
import logging
from os.path import isfile

import zmq

from .models import Keithley2602
from .protocols import CellRunner, STATUS


def main(config):
    """Main start method and loop for server application"""
    # Initialize the channels from a config file
    logging.info("cyckei.server.server.main: Initializing...")

    # Create list of sources (outputs)
    keithleys = []
    sources = []

    # Initialize sources
    logging.info("Attemping {} channels.".format(len(config["channels"])))
    for chd in config["channels"]:
        gpib_addr = chd["gpib_address"]
        keithley = None
        for k in keithleys:
            if gpib_addr == k.gpib_addr:
                keithley = k
        if keithley is None:
            try:
                keithley = Keithley2602(gpib_addr)
            except ValueError as e:
                logging.error("Could not esablish connection: "
                              "Channel {}, GPIB {}.".format(
                                chd["channel"],
                                chd["gpib_address"])
                              )
                logging.error(e)
                continue

            keithleys.append(keithley)

        source = keithley.get_source(chd["keithley_channel"],
                                     channel=chd["channel"])
        sources.append(source)

    logging.info("Connected {} channels.".format(len(sources)))

    # Initialize socket
    runners = []
    socket = initialize_socket(config["zmq"]["port"],
                               config["zmq"]["server"]["address"])

    logging.info(
        "Socket bound to port {}. Entering main loop.".format(
            config["zmq"]["port"])
    )

    max_counter = 1e9
    counter = 0
    initial_time = time.time()
    while True:
        current_time = '{0:02.0f}.{1:02.0f}'.format(
            *divmod((time.time() - initial_time) * 60, 60)
        )

        # check messages on the socket and execute necessary tasks
        # ideally this would happen on a separate thread
        # but it might not be necessary
        # TODO verify this is fast enough to be blocking
        # main loop without problem
        logging.debug("cyckei.server.server.main: Processing socket messages")
        process_socket(socket, runners, sources, current_time)

        # execute runners or sleep if none
        if runners:
            logging.debug("cyckei.server.server.main: Looping over runners")
            # Sort runners from most urgent to least urgent
            runners = sorted(runners, key=lambda x: x.next_time)

            channels_message = ""
            timing_message = ""
            for runner in runners:

                if runner.status == STATUS.pending:
                    runner.run()
                elif runner.status == STATUS.started:
                    # Don't bother with runners more than 1 second out
                    relative_next_time = runner.next_time - time.time()
                    channels_message = channels_message + "/" + runner.channel
                    timing_message = (timing_message
                                      + "/"
                                      + str(relative_next_time)[:5])

                    if relative_next_time <= 0.0:
                        runner.run()
            logging.debug(
                "cyckei.server.server.main: "
                + "channel {} will be checked in approx {} seconds...".format(
                    channels_message[1:], timing_message[1:]
                )
            )

            ipop = sorted(
                [i for i, p in enumerate(runners)
                    if p.status == STATUS.completed],
                reverse=True
            )
            for i in ipop:
                runners.pop(i)

        time.sleep(0.1)

        # mod it by a large value to avoid ever overflowing
        counter = counter % max_counter + 1


def initialize_socket(port, address):
    """Initialize zmq socket"""
    context = zmq.Context(1)
    socket = context.socket(zmq.REP)
    socket.bind("{}:{}".format(address, port))
    return socket


def process_socket(socket, runners, sources, server_time):
    """

    Parameters
    ----------
    socket: zmq.REP socket
        Receives messages in a non-blocking way
        If a message is received it processes it and sends a response

    Returns
    -------
        None

    """

    msg = socket.recv_json()

    if msg is not None:
        response = {"response": None, "message": None}
        try:
            # a message has been received
            fun = msg["function"]
            logging.info("Packet request received: {}".format(fun))
            kwargs = msg.get("kwargs", None)
            # response = {"version": __version__, "response": None}
            resp = "Unknown function"
            if fun == "start":
                try:
                    resp = start(kwargs["channel"], kwargs["meta"],
                                 kwargs["protocol"], runners, sources)
                except Exception:
                    resp = "Error occured when running script."
                    logging.warning("Error occured when running script.")

            elif fun == "pause":
                resp = pause(kwargs["channel"], runners)

            elif fun == "resume":
                resp = resume(kwargs["channel"], runners)

            elif fun == "test":
                resp = test(kwargs["protocol"])

            elif fun == "channel_status":
                resp = channel_status(kwargs["channel"], runners)

            elif fun == "time":
                resp = str(server_time)

            elif fun == "kill_server":
                for protocol in runners:
                    protocol.close()
                response["response"] = "Killing server now"
                socket.send_json(response)
                socket.close()
                socket.context.term()
                exit()

            elif fun == "stop":
                resp = stop(kwargs["channel"], runners)

            elif fun == "ping":
                port = socket.getsockopt_string(
                    zmq.LAST_ENDPOINT
                ).split(":")[-1]
                resp = "True: server is running on port {}".format(port)

            elif fun == "info_all_channels":
                resp = info_all_channels(runners, sources)

            elif fun == "info_channel":
                resp = info_channel(kwargs["channel"], runners)

            response["response"] = resp
            socket.send_json(response)
        except (IndexError, ValueError, TypeError, NameError) as exception:
            response["response"] = (
                "Call failed with error: {}\ntraceback:\n{}".format(
                    exception,
                    traceback.format_exc()
                )
            )
            response["message"] = msg
            logging.debug("cyckei.server.server.process_socket: Sent response")
            socket.send_json(response)


def info_all_channels(runners, sources):
    """Return info on all channels"""
    infos = OrderedDict()
    for source in sources:
        infos[source.channel] = info_channel(source.channel, runners)

    return infos


def info_channel(channel, runners):
    """Return info on specified channels"""
    info = OrderedDict(channel=channel, status=None, state=None,
                       current=None, voltage=None)
    runner = get_runner_by_channel(channel, runners)
    if runner:
        info["status"] = STATUS.string_map[runner.status]
        info["state"] = runner.step.state_str
        data = runner.last_data
        if data:
            info["current"] = data[1]
            info["voltage"] = data[2]
    else:
        info["statuses"] = STATUS.string_map[STATUS.available]

    return info


def start(channel, meta, protocol, runners, sources):
    """Start channel with given protocol"""

    # check to see if there is a already a runner on that channel
    meta["channel"] = channel
    if get_runner_by_channel(channel, runners):
        return "Failed to start channel {}, already in use.".format(channel)

    # check if log file is being used
    path = meta["path"]
    if isfile(path):
        return(
            "Failed to start channel, log file '{}' already in use."
        ).format(path)

    runner = CellRunner(**meta)
    # Set the channel source
    for source in sources:
        if runner.channel == source.channel:
            runner.set_source(source)

    # Protocol must be added after source since some
    # protocol steps require knowledge of the source
    runner.load_protocol(protocol)

    if runner.source is None:
        return "Failed to start channel {}. No source found.".format(channel)

    runners.append(runner)
    return "Succeeded in starting channel {}.".format(channel)


def pause(channel, runners):
    """Pause channel"""
    success = False
    runner = get_runner_by_channel(channel, runners, status=STATUS.started)
    if runner is not None:
        success = runner.pause()

    if success:
        return "Succeeded in pausing channel {}".format(channel)
    return "Failed in pausing channel {}".format(channel)


def stop(channel, runners):
    """Stop channel"""
    success = False
    runner = get_runner_by_channel(channel, runners)
    if runner is not None:
        success = runner.stop()

    if success:
        return "Succeeded in stopping channel {}".format(channel)
    return "Failed in stopping channel {}".format(channel)


def resume(channel, runners):
    """Resume channel from pause"""
    success = False
    runner = get_runner_by_channel(channel, runners, status=STATUS.paused)
    if runner is not None:
        success = runner.resume()

    if success:
        return "Succeeded in resuming channel {}".format(channel)
    return "Failed in resuming channel {}".format(channel)


def test(protocol):
    """Test script load on server"""
    try:
        runner = CellRunner()
        runner.load_protocol(protocol, isTest=True)
        return "Passed"
    except Exception as e:
        return str(e).splitlines()[-1]


def channel_status(channel, runners):
    """Get current state of channel"""
    status = STATUS.string_map[STATUS.available]
    runner = get_runner_by_channel(channel, runners)
    if runner:
        status = STATUS.string_map[runner.status]
    return status


def get_runner_by_channel(channel, runners, status=None):
    """Get runner currently on given channel"""
    if status is None:
        for runner in runners:
            if runner.channel == channel or int(runner.channel) == channel:
                return runner
    else:
        for runner in runners:
            if runner.channel == channel and runner.status == status:
                return runner

    return None



"""Tab to view and edit scripts, also has access to checking procedure"""

from PySide2.QtWidgets import QVBoxLayout, QHBoxLayout, QLabel,\
    QPlainTextEdit, QPushButton, QListWidget, QFileDialog, QMessageBox,\
    QWidget, QStyleOption, QStyle
from PySide2.QtCore import Qt
from PySide2.QtGui import QPainter
from cyckei.client import scripts
from workers import Check


class ScriptEditor(QWidget):
    """Main object of script tab"""
    def __init__(self, channels, scripts):
        QWidget.__init__(self)
        self.channels = channels
        self.scripts = scripts

        # Create overall layout
        columns = QHBoxLayout(self)
        self.setup_file_list()
        columns.addWidget(self.file_list)
        edit_rows = QVBoxLayout()
        columns.addLayout(edit_rows)
        columns.setStretch(0, 1)
        columns.setStretch(1, 5)

        # Create edit_rows
        self.title_bar = QLabel()
        self.title_bar.setText("Select or open file to edit.")
        edit_rows.addWidget(self.title_bar)

        self.editor = QPlainTextEdit()
        self.editor.textChanged.connect(self.text_modified)
        edit_rows.addWidget(self.editor)

        controls = QHBoxLayout()
        edit_rows.addLayout(controls)

        # Create control buttons
        buttons = []
        buttons.append(QPushButton())
        buttons[-1].setText("Open")
        buttons[-1].clicked.connect(self.open)
        buttons.append(QPushButton())
        buttons[-1].setText("Remove")
        buttons[-1].clicked.connect(self.remove)
        buttons.append(QPushButton())
        buttons[-1].setText("New")
        buttons[-1].clicked.connect(self.new)
        buttons.append(QPushButton())
        buttons[-1].setText("Save")
        buttons[-1].clicked.connect(self.save)
        buttons.append(QPushButton())
        buttons[-1].setText("Check")
        buttons[-1].clicked.connect(self.check)
        for button in buttons:
            controls.addWidget(button)

    def setup_file_list(self):
        """Create list of script files"""
        self.file_list = QListWidget()
        self.file_list.itemClicked.connect(self.list_clicked)
        for script in self.scripts.script_list:
            self.file_list.addItem(script)

    def list_clicked(self, item):
        """Display contents of script when clicked"""
        self.title_bar.setText(self.file_list.currentItem().path)
        self.editor.setPlainText(item.content)

    def text_modified(self):
        """Update content of script and update status to show if edited"""
        self.file_list.currentItem().content = self.editor.toPlainText()
        self.file_list.currentItem().update_status()

    def open(self):
        """Open new file and add as script"""
        script_file = QFileDialog.getOpenFileName(
            QWidget(), "Open Script File"
        )[0].rsplit("/", 1)
        if script_file[0]:
            self.add(script_file)

    def remove(self):
        """Remove script from lisr and channel selector"""
        self.scripts.scripts_list.remove(self.file_list.currentItem())
        for channel in self.channels:
            channel.elements[1].removeItem(channel.elements[1].findText(
                    self.file_list.currentItem().title,
                    Qt.MatchFixedString
                ))
        self.file_list.takeItem(self.file_list.currentRow())

    def new(self):
        """Create new file and add to list as script"""
        script_file = QFileDialog.getSaveFileName(QWidget(),
                                                  "Select Directory")[0]
        if script_file:
            open(script_file, "a")
            self.add(script_file.rsplit("/", 1))

    def save(self):
        """Save script"""
        self.file_list.currentItem().save()

    def check(self):
        """Run check protocol to verify validity"""
        if self.threadpool.start(
            Check(scripts.get_script_by_title(
                self.channel.attributes["script_title"]).content)):
            msg = QMessageBox()
            msg.setIcon(QMessageBox.Information)
            msg.setText("Passed!")
            msg.setInformativeText("Script is good to go.")
            msg.setWindowTitle("Check Passed")
            msg.exec_()

    def add(self, file):
        """Add new script to list to make available"""
        self.scripts.script_list.append(scripts.Script(file[1], file[0]))
        self.file_list.addItem(self.scripts.script_list[-1])
        for channel in self.channels:
            channel.elements[1].addItem(self.scripts.script_list[-1].title)

    def paintEvent(self, event):
        style_option = QStyleOption()
        style_option.initFrom(self)
        painter = QPainter(self)
        style = self.style()
        style.drawPrimitive(QStyle.PE_Widget, style_option, painter, self)



import zmq
import json
import logging

from PySide2.QtWidgets import QMessageBox, QWidget

TIMEOUT = 3  # Seconds for listening to server before giving up.


class Socket(object):
    """Handles connection, communication, and control of server over ZMQ"""

    def __init__(self, address, port):
        self.address = address
        self.port = port
        self.socket = zmq.Context().socket(zmq.REQ)
        self.socket.connect("{}:{}".format(self.address, self.port))

    def send(self, to_send):
        """Sends packet to server"""
        logging.debug("Sending: {}".format(to_send["function"]))

        self.socket.send_json(to_send)
        response = self.socket.recv_json()
        logging.debug("Received: {}".format(response))
        return response

    def ping_server(self):
        """Sends ping to server to check if connection functions"""
        ping_response = self.ping()
        msg = QMessageBox()
        msg.setText("Ping response: {}".format(ping_response))
        msg.setWindowTitle("Ping")
        msg.exec_()

    def get_server_time(self):
        """Gets time since server began listening for commands"""
        script = json.loads('{"function": "time", "kwargs": ""}')
        server_time = self.send(script)
        msg = QMessageBox()
        msg.setText("Server time in seconds: {}".format(
            server_time["response"]
        ))
        msg.setWindowTitle("Time")
        msg.exec_()

    def send_file(self, file):
        """Sends packet loaded from JSON file"""
        script = json.load(open(file))
        return self.send(script)

    def channel_status(self, channel):
        """Asks server for "channel_status" information"""
        script = json.loads(
            """{"function": "channel_status", "kwargs": {"channel": """
            + str(channel)
            + """}}""")
        return self.send(script)

    def kill_server(self):
        """Tells server to kill itself and closes connection"""
        reply = QMessageBox.question(
            QWidget(),
            "Message",
            "Kill server?\nAll jobs will be cancelled.",
            QMessageBox.Yes | QMessageBox.No,
            QMessageBox.No
        )

        if reply == QMessageBox.Yes:
            script = json.loads('{"function": "kill_server"}')
            print(self.send(script))
            self.close_socket()
        else:
            pass

    def ping(self):
        """Send "ping" to server"""
        script = json.loads('{"function": "ping"}')
        response = self.send(script)["response"]
        return response

    def info_channel(self, channel):
        """Send "info_channel" to server"""
        script = json.loads(
            """{"function": "info_channel", "kwargs": {"channel": """
            + str(channel)
            + """}}""")
        return self.send(script)

    def info_all_channels(self):
        """Send "info_all_channels" to server"""
        script = json.loads('{"function": "info_all_channels"}')
        return self.send(script)



import logging

from pkg_resources import require, DistributionNotFound
from json import load
from os.path import expanduser, exists
from os import makedirs
from shutil import copy
from subprocess import Popen

from cyckei.server import server


def main():
    """
        This is the main entrypoint and routine of Cyckei Vayu.
        Vayu aims to make cyckei more efficient in many aspects.
    """
    print("\n")

    try:
        version = require("cyckei")[0].version
    except DistributionNotFound:
        version = "(unpackaged)"

    print("Welcome to Cyckei Vayu version {}.".format(version))

    # Setup recording direcory if unavailable
    print("Checking for recording directory...")
    record_dir = expanduser("~") + "/cyckei"
    file_structure(record_dir)

    # Load configuration
    print("Loading configuration...")
    config = load(open(record_dir + "/config.json", "r"))
    config["path"] = record_dir

    # Start logging
    log_file = "{}/server.log".format(record_dir)
    print("Starting log...")
    logging.basicConfig(filename=log_file, level=config["verbosity"],
                        format='%(asctime)s %(message)s')
    logging.info("--- Server started.")

    # Open client window
    Popen(["python", "./cyckei/client/client.py"])

    # Start server
    server.main(config)


def file_structure(path):
    """Checks for existing folder structure and sets up if missing"""

    if not exists(path):
        makedirs(path)
    if not exists(path + "/tests"):
        makedirs(path + "/tests")
    if not exists(path + "/config.json"):
        copy("resources/default.config.json", path + "/config.json")
    if not exists(path + "/batch.txt"):
        copy("resources/batch.txt", path + "/batch.txt")
    if not exists(path + "/scripts"):
        makedirs(path + "/scripts")
        copy("resources/example-script", path + "/scripts/example")


if __name__ == "__main__":
    main()



import json
import time
from datetime import datetime
import operator
import logging

DATETIME_FORMAT = '%Y-%m-%d_%H:%M:%S.%f'
NEVER = 2 * time.time()

OPERATOR_MAP = {
    "<": operator.lt,
    "<=": operator.le,
    "=<": operator.le,
    "==": operator.eq,
    "!=": operator.ne,
    "=!": operator.ne,
    ">=": operator.ge,
    "=>": operator.ge,
    ">": operator.gt,
    "lt": operator.lt,
    "le": operator.le,
    "eq": operator.eq,
    "ne": operator.ne,
    "ge": operator.ge,
    "gt": operator.gt}

DATA_INDEX_MAP = {
    "time": 0,
    "current": 1,
    "voltage": 2,
    "capacity": 3
}

DATA_NAME_MAP = {v: k for (k, v) in DATA_INDEX_MAP.items()}


class VariableHolder:
    pass


STATUS = VariableHolder()
STATUS.available = -1
STATUS.pending = 0
STATUS.started = 1
STATUS.paused = 2
STATUS.completed = 3
STATUS.unknown = 4
STATUS.string_map = {
    STATUS.pending: "pending",
    STATUS.started: "started",
    STATUS.paused: "paused",
    STATUS.completed: "completed",
    STATUS.unknown: "unknown",
    STATUS.available: "available"
}


class CellRunner(object):
    META = {
        "path": None,
        "protocol_name": None,
        "protocol": None,
        "channel": None,
        "cellid": None,
        "comment": None,
        "requester": None,
        "cycler": None,
        "cell_type": None,
        "package": None,
        "start_cycle": None
    }

    def __init__(self, **meta):
        self.meta = self.META.copy()
        for k in self.meta.keys():
            self.meta[k] = meta.get(k, None)

        if self.meta["cycler"] is None:
            self.meta["cycler"] = "Keithley2602"
        if self.meta["cell_type"] is None:
            self.meta["cell_type"] = "unknown"

        # Enforce a str channel
        self.meta["channel"] = str(self.meta["channel"])
        self.fpath = self.meta["path"]
        self.steps = []
        self.status = STATUS.pending
        self.current_step = None
        self.start_time = None
        self.i_current_step = None
        if self.meta["start_cycle"] is not None:
            self.cycle = self.meta["start_cycle"]
        else:
            self.cycle = 0
        self.prev_cycle = None
        self.next_time = -1
        self.channel = self.meta["channel"]
        self.data_format = "    " + ",".join(["{:0.8g}"] * 4) + "\n"
        self.last_data = None
        self.total_pause_time = 0.0
        self.source = None

    def set_source(self, source):
        self.source = source
        if self.channel != source.channel:
            raise ValueError(
                "the runner channel ({}) and the source \
                    channel ({}) should be identical".format(
                        self.channel, source.channel
                    ))

    def set_cap_signs(self, direction=None):
        """
        Go through the steps and set their .cap_sign attribute
        based on the "direction" of teh cell.
        This decides whether charging yields positive or negative capacities
        and vice versa for discharge.

        Parameters
        ----------
        direction: str or None
            direction for the cell can be "pos", "neg", or None
            if it is None, the "cell_type" in teh meta data is used,
            if this is not set it defaults to "pos"


        Returns
        -------
        None
        """
        valid_directions = ["pos", "neg", None]
        if direction not in valid_directions:
            raise ValueError(
                "received unsupported direction argument:\
                {}, must be one of {}".format(direction, valid_directions))

        if direction is None:
            direction = 'pos'
            if "anode" in self.meta["cell_type"].lower():
                direction = 'neg'

        if direction == "pos":
            for step in self.steps:
                if step.state_str.startswith("charge"):
                    step.cap_sign = 1.0
                elif step.state_str.startswith("discharge"):
                    step.cap_sign = -1.0
                else:
                    step.cap_sign = 1.0

        elif direction == "neg":
            for step in self.steps:
                if step.state_str.startswith("charge"):
                    step.cap_sign = -1.0
                elif step.state_str.startswith("discharge"):
                    step.cap_sign = 1.0
                else:
                    step.cap_sign = 1.0

    def load_protocol(self, protocol: str, isTest=False):
        """

        Parameters
        ----------
        protocol: str
            string of python code generating the protocol steps in the runner

        Returns
        -------

        """
        # The protocol is a python string that is executed as code.
        # The CellRunner instance must be present as "parent" in the globals
        # so that the ProtocolSteps can add themselves to the .steps list
        self.isTest = isTest
        exec(protocol, globals().update({"parent": self}))
        # Set the signs for the capacity calculations
        self.set_cap_signs()

    def add_step(self, step):
        self.steps.append(step)

    def next_step(self):
        self.i_current_step += 1
        if self.i_current_step >= len(self.steps):
            # reached end of the steps
            self.status = STATUS.completed
            return False

        # Give the last recorded capacity as a starting point to the new step
        if self.last_data is not None:
            self.step.starting_capacity = self.last_data[3]

        # On a new step you want to start as soon as possible on the main loop
        self.next_time = time.time()
        return True

    def _start(self):
        """
        Initializes the protocol, would not normally be called directly,
        but instead is called by the .run() function

        Returns
        -------
        None
        """
        logging.info("cyckei.server.protocols.CellRunner._start: \
            Starting cellrunner instance (channel: {})".format(self.channel))

        self.i_current_step = 0
        self.status = STATUS.started
        self.start_time = time.time()
        self.meta["date_start_timestamp"] = self.start_time
        self.meta["date_start_timestr"] = (
            datetime.fromtimestamp(self.start_time).strftime(DATETIME_FORMAT)
        )
        self.write_header()
        self.write_cycle_header()

    @property
    def step(self):
        try:
            return self.steps[self.i_current_step]
        except IndexError:
            return None

            # TODO should next_time be a property?
            # currently there is a protocol.next_time and a step.next_time.
            # TODO Not sure if this is redundant
            #    @property
            #    def next_time(self):
            #        return self.step.next_time

    def run(self, force_report=False):
        """
        Method called by the main loop to start and advance a protocol
        This method triggers the various steps that are loaded in the protocol
        as well as the header writing and the data writing in the data file

        Returns
        -------
        success: bool
            True if is running as expected
            False if it is complete

        """
        logging.debug("cyckei.server.protocols.CellRunner.run: \
            Entering method for channel {}".format(self.channel))
        if self.status == STATUS.completed:
            return False

        if self.status == STATUS.paused:
            return False

        if self.status != STATUS.started:
            self._start()

        if self.step.status != STATUS.started:
            self.write_step_header()

        self.read_and_write(force_report=force_report)

        if self.step.status == STATUS.completed:
            self.next_step()

        if self.status == STATUS.completed:
            self.close()

        return True

    def advance_cycle(self):
        self.cycle += 1
        self.write_cycle_header()

    def write_header(self):
        with open(self.fpath, 'w') as fo:
            header = json.dumps(self.meta, indent=4, sort_keys=True)
            header = "\n".join(
                ["#{}".format(line) for line in header.split("\n")]
            )
            fo.write(header + "\n")

    def write_cycle_header(self):
        with open(self.fpath, 'a') as fo:
            cycle_header = json.dumps({"cycle": self.cycle}) + "\n"
            fo.write(cycle_header)

    def write_step_header(self):
        header = self.step.header()
        if header:
            with open(self.fpath, 'a') as fo:
                fo.write("  " + header + "\n")

    def read_and_write(self, force_report=False):
        data = self.step.run(force_report=force_report)
        self.next_time = self.step.next_time
        if data:
            self.write_data(*data)
            self.last_data = data

    def write_data(self, timestamp, current, voltage, capacity):
        with open(self.fpath, 'a') as fo:
            fo.write(self.data_format.format(
                timestamp - self.start_time - self.total_pause_time,
                current, voltage, capacity)
            )

    def pause(self):
        """
        User must be able to pause a cell arbitrarily
        Returns
        -------

        """
        success = False
        self.read_and_write(force_report=True)
        self.status = STATUS.paused
        if self.step.pause():
            self.next_time = self.step.next_time
            success = True
        return success

    def resume(self):
        self.status = STATUS.started
        self.step.resume()
        return self.run()

    def close(self):
        """
        This is to close
        Returns
        -------

        """
        self.source.off()

    def off(self):
        """
        This is just to turn off the current
        Returns
        -------

        """
        self.source.off()

    def stop(self):
        self.status = STATUS.completed
        self.close()
        return True


class ProtocolStep(object):
    """
    Base class for a protocol step, needs to be subclassed with
    implementation of a start function

    A protocol step stores stores its own data and the reported points
    Keeps track of time, current, voltage, capacity

    Because the protocol files are simply pure python the parent, which is a
    CellRunner instance needs to be present in the global variables as "parent"

    """

    def __init__(self,
                 wait_time: float = 10.0,
                 cellrunner_parent: CellRunner = None):
        """
        Base class for protocols
        the variable "parent" must be a CellRunner
        instance and be present in the globals during instantiation

        Parameters
        ----------
        wait_time: float
            default waiting time in seconds
        """
        # the parent is the CellRunner
        if cellrunner_parent is None:
            self.parent = parent
        else:
            self.parent = cellrunner_parent

        # list of lists [[time,current,voltage,capacity],
        # [time,current,voltage,capacity], ...]
        # one entry for each measurement, current is stored in absolute value
        self.data = []
        self.data_max_len = 10000
        # same format as data but only for the reported points
        self.report = []

        # Status must be one of the values in STATUS module variable
        self.status = STATUS.pending

        # State string, defaults to unknown
        self.state_str = "unknown"

        # Never run before
        self.last_time = -1

        # Need to track time spent in pause state for time-based conditions
        self.pause_start = None
        self.pause_time = 0.0

        # The cap sign determines whether the capacitiy
        # increases or decreases during
        # charge and discharge. The parent (protocol) must set this
        # attribute with .set_cap_sign() since
        # it depends on the cell type and is unknown to the step itself
        self.cap_sign = 1.0

        # this is the time in seconds since epoch at which this protocol
        # step is expecting to do another read
        # operation on the channel. -1 means it has never been set.
        self.next_time = -1

        # This gets set at the protocol level (parent)
        self.starting_capacity = 0.

        # Wait time is defaults wait time between data measurements
        # assumming no other conditions are present
        self.wait_time = wait_time

        # Lists which will hold the conditions for reporting and ending
        self.end_conditions = []
        self.report_conditions = []

        self.parent.add_step(self)

    def _start(self):
        raise NotImplementedError

    def header(self):
        return False

    def run(self, force_report=False):
        """

        Parameters
        ----------
        force_report: bool
            defaults to False. The run function evaluates the
            report_conditions to decide if a data point should be
            reported, setting force_report to True will report the
            latest data point regardless of conditions.


        Returns
        -------
        data_to_report: tuple of 4
            (time, current, voltage, capacity) tuple to report (write to file)
            returns non if no data to report

        """
        logging.debug(
            "cyckei.server.protocols.ProtocolStep.run: \
            running {} protocol on channel {}".format(self.state_str,
                                                      self.parent.channel))
        if self.status == STATUS.paused:
            self.next_time = NEVER
            return None

        if self.status != STATUS.started:
            self._start()

        self.read_data()

        # Set the next read time using the default wait_time
        # this may get modified by the evaluations of conditions
        self.next_time = self.data[-1][0] + self.wait_time

        self.check_end_conditions()

        if self.status == STATUS.completed:
            report_data = True

        else:
            report_data = self.check_report_conditions()

        if report_data or force_report:
            self.report.append(self.data[-1])
            return self.report[-1]
        else:
            return None

    def check_end_conditions(self):
        """

        Returns
        -------
        bool: whether one of the end conditions was satisfied
        """
        for condition in self.end_conditions:
            if condition.check(self):
                self.status = STATUS.completed
                return True
        return False

    def check_report_conditions(self):
        """

        Returns
        -------
        bool: whether one of the report conditions was satisfied
        """
        for condition in self.report_conditions:
            if condition.check(self):
                return True
        return False

    def read_data(self, force_report=False):
        self.last_time = time.time()
        current, voltage = self.parent.source.read_iv()

        # Currents are reported only in absolute values
        current = abs(current)

        if self.data:
            # Record the capacity in mAh
            capacity = (self.data[-1][3]
                        + (self.last_time - self.data[-1][0])
                        / 3600.
                        * self.cap_sign
                        * (current + self.data[-1][1])
                        / 2.0
                        / 1000.)
        else:
            capacity = self.starting_capacity

        self.data.append([self.last_time, current, voltage, capacity])

        if len(self.data) > self.data_max_len:
            # we pop 1 and not 0
            # because if we pop 0 it will screw up the total time checking
            self.data.pop(1)

        if force_report:
            self.report.append(self.data[-1])

    def pause(self):
        if self.status != STATUS.started:
            return False

        self.status = STATUS.paused
        self.pause_start = time.time()
        self.parent.off()
        self.next_time = NEVER
        return True

    def resume(self):
        if self.status != STATUS.paused:
            return False
        self.pause_time = time.time() - self.pause_start
        self.parent.total_pause_time += self.pause_time
        self._start()
        # self.status = STATUS.started


def process_reports(reports):
    """

    Parameters
    ----------
    reports:
        desired reports as a tuple of pairs such as
        (("voltage",0.01), ("time",300))

    Returns
    -------
    report_conditions: list
        List of condition objects for reporting a data point

    """
    report_conditions = []
    for k, v in reports:
        if k == "voltage":
            report_conditions.append(condition_dv(v))
        elif k == "time":
            report_conditions.append(condition_dt(time_conversion(v)))
        elif k == "current":
            report_conditions.append(condition_di(v))
        elif k == "capacity":
            report_conditions.append(condition_dc(v))
    return report_conditions


def process_ends(ends):
    """

    Parameters
    ----------
    ends:
        desired ends conditions as a tuple of triples such as
        (("voltage",">", 4.2), ("time",">","24::"))

    Returns
    -------
    end_conditions: list
        List of condition objects for ending a protocol step

    """
    end_conditions = []
    for args in ends:
        if args[0] == "time":
            time_seconds = time_conversion(args[2])
            end_conditions.append(ConditionTotalTime(time_seconds))
        else:
            end_conditions.append(ConditionAbsolute(*args))
    return end_conditions


class CurrentStep(ProtocolStep):
    def __init__(self, current,
                 reports=(("voltage", 0.01), ("time", ":5:")),
                 ends=(("voltage", ">", 4.2), ("time", ">", "24::")),
                 wait_time=10.0):
        super().__init__(wait_time=wait_time)

        if current > 0:
            self.state_str = "charge_constant_current"
            sign = 1
        elif current < 0:
            self.state_str = "discharge_constant_current"
            sign = -1
        else:
            raise ValueError("current argument should be non-zero")
        self.current = current

        self.v_limit = sign * 5.0
        for e in ends:
            if e[0] == "voltage":
                self.v_limit = e[2] + sign * 0.1

        self.report_conditions = process_reports(reports)
        self.end_conditions = process_ends(ends)

    def _start(self):
        self.status = STATUS.started
        self.parent.source.set_current(current=self.current,
                                       v_limit=self.v_limit)

    def header(self):
        return json.dumps({"state": self.state_str,
                          "date_start_timestr": datetime.now().strftime(
                                DATETIME_FORMAT)}
                          )


class CCCharge(CurrentStep):
    def __init__(self, current,
                 reports=(("voltage", 0.01), ("time", ":5:")),
                 ends=(("voltage", ">", 4.2), ("time", ">", "24::")),
                 wait_time=10.0):
        # Enforce positive current
        current = abs(current)
        super().__init__(current,
                         reports=reports, ends=ends,
                         wait_time=wait_time)
        self.state_str = "charge_constant_current"


class CCDischarge(CurrentStep):
    def __init__(self, current,
                 reports=(("voltage", 0.01), ("time", ":5:")),
                 ends=(("voltage", "<", 3), ("time", ">", "24::")),
                 wait_time=10.0):
        # Enforce negative current
        current = -abs(current)
        super().__init__(current,
                         reports=reports, ends=ends,
                         wait_time=wait_time)
        self.state_str = "discharge_constant_current"


class VoltageStep(ProtocolStep):
    def __init__(self, voltage,
                 reports=(("current", 0.01), ("time", ":5:")),
                 ends=(("current", "<", 0.001), ("time", ">", "24::")),
                 wait_time=10.0):
        super().__init__(wait_time=wait_time)

        self.i_limit = None
        self.voltage = voltage
        self.report_conditions = process_reports(reports)
        self.end_conditions = process_ends(ends)

    def guess_i_limit(self):
        self.i_limit = 1.0
        if self.state_str.startswith("charge"):
            self.i_limit = self.parent.source.current_ranges[-1]
        elif self.state_str.startswith("discharge"):
            self.i_limit = -self.parent.source.current_ranges[-1]

    def _start(self):
        self.status = STATUS.started
        if self.i_limit is None:
            raise ValueError(
                "the i_limit must be set, try using the guess_i_limit method"
            )

        self.parent.source.set_voltage(voltage=self.voltage,
                                       i_limit=self.i_limit)

    def header(self):
        return json.dumps({"state": self.state_str,
                          "date_start_timestr": datetime.now().strftime(
                            DATETIME_FORMAT)}
                          )


class CVCharge(VoltageStep):
    def __init__(self, voltage,
                 reports=(("current", 0.01), ("time", ":5:")),
                 ends=(("current", "<", 0.001), ("time", ">", "24::")),
                 wait_time=10.0):
        super().__init__(voltage,
                         reports=reports, ends=ends,
                         wait_time=wait_time)
        self.state_str = "charge_constant_voltage"
        if parent.isTest:
            pass
        else:
            self.guess_i_limit()


class CVDischarge(VoltageStep):
    def __init__(self, voltage,
                 reports=(("current", 0.01), ("time", ":5:")),
                 ends=(("current", "<", 0.001), ("time", ">", "24::")),
                 wait_time=10.0):
        super().__init__(voltage,
                         reports=reports, ends=ends,
                         wait_time=wait_time)
        self.state_str = "discharge_constant_voltage"
        if parent.isTest:
            pass
        else:
            self.guess_i_limit()


class AdvanceCycle(ProtocolStep):
    def _start(self):
        self.status = STATUS.started

    def run(self, force_report=False):
        self._start()
        self.parent.advance_cycle()
        self.status = STATUS.completed


class Rest(ProtocolStep):
    def __init__(self,
                 reports=(("time", ":5:"),), ends=(("time", ">", "24::"),),
                 wait_time=10.0):
        super().__init__(wait_time=wait_time)

        self.state_str = "rest"
        self.report_conditions = process_reports(reports)
        self.end_conditions = process_ends(ends)

    def _start(self):
        self.status = STATUS.started
        self.parent.source.rest()

    def header(self):
        return json.dumps({"state": self.state_str,
                          "date_start_timestr": datetime.now().strftime(
                            DATETIME_FORMAT)}
                          )


class Pause(ProtocolStep):
    def __init__(self):
        """

        Parameters
        ----------
        parent: Protocol
            Protocol to which this protocol step belongs

        """
        # This is the time between measurements on the channel,
        # putting this arbitrarily large
        super().__init__(wait_time=NEVER)
        self.state_str = "pause"

    def _start(self):
        self.state = STATUS.started
        self.next_time = time.time() + self.wait_time
        self.parent.source.pause()

    def run(self):
        self._start()
        return None

    def resume(self):
        self.status = STATUS.completed


class Sleep(ProtocolStep):
    def __init__(self,
                 reports=(("time", ":5:"),), ends=(("time", ">", "24::"),),
                 wait_time=10.0):
        super().__init__(wait_time=wait_time)

        self.state_str = "sleep"
        self.report_conditions = process_reports(reports)
        self.end_conditions = process_ends(ends)

    def _start(self):
        self.status = STATUS.started
        self.parent.source.off()

    def header(self):
        return json.dumps({"state": self.state_str,
                          "date_start_timestr": datetime.now().strftime(
                            DATETIME_FORMAT)}
                          )

    def read_data(self):
        self.parent.source.rest()
        super().read_data()
        self.parent.source.off()

    def run(self, force_report=False):
        """
        The run method needs to be redefined because the logic of the Sleep
        protocol is unique in that a measurement is
        only desired if a report condition is met as opposed to
        the other steps which measure and then decide whether to report

        Parameters
        ----------
        force_report: bool
            defaults to False. The run function evaluates the
            report_conditions to decide if a data point should be
            reported, setting force_report to True will report the
            latest data point regardless of conditions.


        Returns
        -------
        data_to_report: tuple of 4
            (time, current, voltage, capacity) tuple to report (write to file)
            returns none if no data to report

        """

        if self.status == STATUS.paused:
            self.next_time = NEVER
            return None

        report_data = False
        if self.status != STATUS.started:
            self._start()
            report_data = True

        self.next_time = time.time() + self.wait_time

        self.check_end_conditions()

        if self.status == STATUS.completed:
            report_data = True

        else:
            report_data = report_data or self.check_report_conditions()

        if report_data or force_report:
            self.read_data()
            self.report.append(self.data[-1])
            return self.report[-1]
        else:
            return None


class Condition(object):
    """
    A Condition is an object which given a ProtocolStep
    object to its "check" method will return a boolean
    indicating whether that condition has been met

    A condition object may modify the .next_time attribute of
    the ProtocolStep object to suggest time the
    next time as an absolute timestamp that the condition should be checked
    """

    def check(self, protocol_step: ProtocolStep):
        raise NotImplementedError


class ConditionDelta(Condition):
    def __init__(self, value_str: str, delta: float):
        """
        Change between latest reported value and latest measured value

        Parameters
        ----------
        value_str: str
            value string such as "voltage", "time", "current" etc...
            See DATA_INDEX_MAP module variable
            for valid values
        delta: float
            Delta to compare against
        """
        self.value_str = value_str
        self.index = DATA_INDEX_MAP[value_str]
        self.comparison = OPERATOR_MAP[">="]
        self.delta = abs(delta)
        self.is_time = value_str == "time"

    def check(self, step):
        try:
            if len(step.report) == 0:
                step.next_time = time.time()
                return True
            else:
                if self.is_time:
                    val = time.time()
                    delta = (abs(val - step.report[-1][self.index])
                             - step.pause_time)
                    next_time = self.delta - delta + val
                    step.next_time = min(next_time, step.next_time)
                else:
                    val = step.data[-1][self.index]
                    next_time = extrapolate_time(step.data,
                                                 val + self.delta,
                                                 self.index)
                    step.next_time = min(next_time, step.next_time)

                logging.debug("cyckei.server.protocols.ConditionDelta: \
                {}, set next_time to {}".format(self.value_str,
                                                step.next_time))

                if self.comparison(abs(val - step.report[-1][self.index]),
                                   self.delta):
                    return True
                else:
                    return False
        except IndexError:
            return False


class ConditionTotalDelta(Condition):
    def __init__(self, value_str: str, delta: float):
        """
        Change between first reported value and latest measured value

        Parameters
        ----------
        value_str: str
            value string such as "voltage", "time", "current" etc...
            See DATA_INDEX_MAP module variable
            for valid values
        delta: float
            Delta to compare against
        """
        self.delta = delta
        self.value_str = value_str
        self.index = DATA_INDEX_MAP[value_str]
        self.comparison = OPERATOR_MAP[">="]
        # at what timestamp do we expect the condition to be met
        self.next_time = NEVER

    def check(self, step):
        try:
            delta = abs(step.data[-1][self.index] - step.report[0][self.index])
            return self.comparison(delta, self.delta)
        except IndexError:
            return False


class ConditionTotalTime(ConditionTotalDelta):
    def __init__(self, delta):
        """
        Change between first reported time and latest reported time

        Parameters
        ----------
        delta: float
            Total elapsed time in seconds
        """

        super().__init__("time", delta)

    def check(self, step):
        try:
            # Previously was using the last step.data to do the check, however,
            # in the case where no measurements are
            # performed prioir to evaluating the condition this does not give
            # the correct result
            # I see no downside to taking the clock time instead
            delta = abs(time.time() - step.report[0][self.index]) \
                    - step.pause_time

            if self.comparison(delta, self.delta):
                return True
            else:
                next_time = self.delta - delta + time.time()
                step.next_time = min(step.next_time, next_time)
                logging.debug(
                    "cyckei.server.protocols.ConditionTotalTime: \
                    {}, set next_time to {}".format(self.value_str,
                                                    step.next_time))
                return False
        except IndexError:
            return False


class ConditionAbsolute(Condition):
    def __init__(self, value_str: str, operator_str: str,
                 value: float, min_time: float = 1.0):
        """

        Parameters
        ----------
        value_str: str
            value string such as "voltage", "time", "current" etc...
            See DATA_INDEX_MAP module variable
            for valid values
        operator_str: str
            String indicating the ocmparison operator.
            See OPERATOR_MAP module variable for valid values
        value: float
            Actual value to compare against
        min_time: float
            minimum time that must have elapsed before evaluating the condition
        """
        self.value = value
        self.value_str = value_str
        self.index = DATA_INDEX_MAP[value_str]
        self.comparison = OPERATOR_MAP[operator_str]
        # at what timestamp do we expect the condition to be met
        self.next_time = 1e30
        self.min_time = min_time

    def check(self, step):
        try:
            execute_check = True
            if self.min_time is not None:
                execute_check = (
                    step.data[-1][0] - step.report[0][0] > self.min_time
                )

            if execute_check:
                if self.comparison(step.data[-1][self.index], self.value):
                    return True
                else:
                    next_time = extrapolate_time(step.data,
                                                 self.value,
                                                 self.index)
                    step.next_time = min(next_time, step.next_time)
                    logging.debug(
                        "cyckei.server.protocols.ConditionTotalTime: \
                        {}, set next_time to {}".format(self.value_str,
                                                        step.next_time))
                    return False
            else:
                return False
        except IndexError:
            return False


def condition_end_voltage(voltage, operator_str):
    return ConditionAbsolute("voltage", operator_str, voltage)


def condition_ucv(voltage):
    return ConditionAbsolute("voltage", ">=", voltage)


def condition_lcv(voltage):
    return ConditionAbsolute("voltage", "<=", voltage)


def condition_min_current(current):
    return ConditionAbsolute("current", "<=", current)


def condition_max_current(current):
    return ConditionAbsolute("current", ">=", current)


def condition_total_time(total_time):
    return ConditionTotalTime(total_time)


def condition_dt(dt):
    return ConditionDelta("time", dt)


def condition_di(di):
    return ConditionDelta("current", di)


def condition_dv(dv):
    return ConditionDelta("voltage", dv)


def condition_dc(dc):
    return ConditionDelta("capacity", dc)


def extrapolate_time(data, target, index):
    try:
        # Project when it will be hit
        d1 = data[-1]
        # require at least 10 mV of difference for projection
        for i in range(2, 100):
            d0 = data[-i]
            if abs(d1[2] - d0[2]) > 0.0001:
                break
        next_time = ((target - d1[index])
                     / (d1[index] - d0[index])
                     * (d1[0] - d0[0]) + d1[0])

    except (NameError, IndexError, ZeroDivisionError):
        next_time = time.time()

    logging.debug(
        "cyckei.server.protocols.extrapolate_time: \
        Extrapolated time {} using {} index and target value {}".format(
            next_time, DATA_NAME_MAP[index], target)
        )

    return next_time


def time_conversion(t):
    """

    Parameters
    ----------
    t: str or float
        time in the "hh:mm:ss" format, where values can be ommitted
        e.g. "::5" would be five seconds or
        time in seconds

    Returns
    -------
    time (float) in seconds
    """
    t_float = None
    try:
        t_float = float(t)
    except ValueError:
        try:
            time_tuple = [float(x) if x else 0 for x in t.split(":")]
            t_float = (time_tuple[0]
                       * 3600.
                       + time_tuple[1]
                       * 60.0
                       + time_tuple[2])
        except AttributeError:
            pass
    return t_float



"""Controls log tab, which displays logs as they are being recorded"""

from os import listdir, path
import subprocess
from PySide2.QtWidgets import QVBoxLayout, QHBoxLayout, QLabel, QPlainTextEdit,\
    QPushButton, QListWidget, QListWidgetItem, QWidget, QStyleOption, QStyle
from PySide2.QtGui import QPainter


class LogViewer(QWidget):
    """Object of log tab"""
    def __init__(self, config):
        QWidget.__init__(self)
        self.path = config["path"] + "/tests"

        self.file_list = QListWidget()
        self.file_list.itemClicked.connect(self.list_clicked)

        # Create overall layout
        columns = QHBoxLayout(self)
        columns.addWidget(self.file_list)
        edit_rows = QVBoxLayout()
        columns.addLayout(edit_rows)
        columns.setStretch(0, 1)
        columns.setStretch(1, 5)

        # create edit_rows
        self.title_bar = QLabel()
        self.title_bar.setText("Select file on left to view log.")
        edit_rows.addWidget(self.title_bar)

        self.editor = QPlainTextEdit()
        edit_rows.addWidget(self.editor)

        controls = QHBoxLayout()
        edit_rows.addLayout(controls)

        # create control buttons
        reload = QPushButton()
        reload.setText("Reload")
        reload.clicked.connect(self.reload)
        controls.addWidget(reload)

        open_folder = QPushButton()
        open_folder.setText("Open Folder")
        open_folder.clicked.connect(self.open_explorer)
        controls.addWidget(open_folder)

        self.reload()

    def reload(self):
        """Reload list and contents of logs from specified logging folder"""
        self.file_list.clear()
        self.editor.clear()
        logs = []

        files = listdir(self.path)
        if files is not None:
            for file in files:
                if path.isdir(path.join(self.path, file)):
                    pass
                else:
                    logs.append(Log(file, self.path))

        for log in logs:
            self.file_list.addItem(log)

    def open_explorer(self):
        """Open logging folder in explorer"""
        subprocess.Popen(r'explorer /select,"{}"'.format(self.path))

    def list_clicked(self, item):
        """Display text of clicked file in text box"""
        self.editor.setPlainText(item.content)

    def paintEvent(self, event):
        style_option = QStyleOption()
        style_option.initFrom(self)
        painter = QPainter(self)
        style = self.style()
        style.drawPrimitive(QStyle.PE_Widget, style_option, painter, self)


class Log(QListWidgetItem):
    """Object of log, stores title and content of file for quick access"""
    def __init__(self, title, file_path):
        super().__init__()
        self.title = title
        self.setText(self.title)
        self.content = open(file_path + "/" + self.title, "r").read()



"""
Methods that controls the cykei client.
Controls communication and initializes the MainWindow.
"""

import json
import sys
import logging

from PySide2.QtWidgets import QApplication
from os.path import expanduser

from cyckei.client.window import MainWindow


def main():
    """Initializes server and window"""

    # Load configuration
    record_dir = expanduser("~") + "/cyckei"
    config = json.load(open(record_dir + "/config.json", "r"))
    config["path"] = record_dir

    # Start logging
    log_file = "{}/client.log".format(record_dir)
    logging.basicConfig(filename=log_file, level=config["verbosity"],
                        format='%(asctime)s %(message)s')
    logging.info("--- Client started.")

    app = QApplication(sys.argv)
    window = MainWindow(config)
    window.show()

    sys.exit(app.exec_())


if __name__ == "__main__":
    main()



"""
Widget that controls a single channnel.
Listed in the channel tab of the main window.
"""

import json
import logging


from PySide2.QtWidgets import QWidget, QVBoxLayout, QHBoxLayout, QComboBox, \
    QLineEdit, QPushButton, QLabel, QScrollArea, QStyleOption, \
    QStyle
from PySide2.QtCore import QMetaObject
from PySide2.QtGui import QPainter

import workers


def not_none(value):
    """Sets a None value to "None" string"""
    return "None" if value is None else str(value)


def new_button_element(text, status, connect):
    """Creates a button with given information"""
    button = QPushButton()
    button.setText(text)
    button.setStatusTip(status)
    button.clicked.connect(connect)
    return button


def new_combo_element(items, status, connect):
    """Creates a combo box with given information"""
    combo_box = QComboBox()
    combo_box.addItems(items)
    combo_box.setStatusTip(status)
    combo_box.activated[str].connect(connect)
    return combo_box


def new_text_element(label, status, connect):
    """Creates text edit field with given information"""
    text = QLineEdit()
    text.setPlaceholderText(label)
    text.setStatusTip(status)
    text.textChanged.connect(connect)
    return text


class ChannelTab(QWidget):
    def __init__(self, config, threadpool, scripts):
        """Setup each channel widget and place in QVBoxlayout"""
        QWidget.__init__(self)

        layout = QHBoxLayout(self)

        scroll_area = QScrollArea()
        layout.addWidget(scroll_area)
        scroll_area.setWidgetResizable(True)
        scroll_contents = QWidget()
        scroll_area.setWidget(scroll_contents)
        rows = QVBoxLayout(scroll_contents)
        rows.setSpacing(0)

        self.channels = []
        for channel in config["channels"]:
            self.channels.append(ChannelWidget(
                channel["channel"],
                config["path"] + "/tests",
                threadpool,
                scripts
            ))
            rows.addWidget(self.channels[-1])

    def paintEvent(self, event):
        style_option = QStyleOption()
        style_option.initFrom(self)
        painter = QPainter(self)
        style = self.style()
        style.drawPrimitive(QStyle.PE_Widget, style_option, painter, self)


class ChannelWidget(QWidget):
    """Controls and stores information for a given channel"""

    def __init__(self, channel, record_folder, threadpool, scripts):
        super().__init__()
        # Default Values
        self.attributes = {}
        self.attributes["channel"] = channel
        self.attributes["record_folder"] = record_folder
        self.attributes["id"] = 0
        self.attributes["comment"] = "No Comment"
        self.attributes["package"] = "Pouch"
        self.attributes["type"] = "Full"
        self.attributes["path"] = "default.pyb"
        self.attributes["mass"] = 1
        self.attributes["requestor"] = "Unspecified"
        self.attributes["protocol_name"] = None
        self.attributes["script_title"] = None

        self.threadpool = threadpool
        self.scripts = scripts

        self.setMinimumSize(800, 54)

        # Setup UI
        rows = QVBoxLayout(self)
        settings = QHBoxLayout()
        rows.addLayout(settings)
        self.setup_ui()

        for element in self.elements:
            settings.addWidget(element)
        rows.addWidget(self.elements[-1])
        if (int(self.attributes["channel"]) % 2 == 0):
            self.setObjectName("even")
        else:
            self.setObjectName("odd")

        # Load default JSON
        self.json = json.load(open("resources/defaultJSON.json"))

        # Update status
        self.threadpool.start(workers.UpdateStatus(self))

    def setup_ui(self):
        """Creates all UI elements and adds them to self.elements list"""
        self.elements = []

        # 0 - Cell channel label
        self.elements.append(QLabel())
        self.elements[-1].setStatusTip(
            "Channel {}".format(self.attributes["channel"])
        )
        self.elements[-1].setText(
            "{}:".format(self.attributes["channel"])
        )
        self.elements[-1].setObjectName("id_label")

        # 1 - Script selection box
        available_scripts = []
        if self.scripts.script_list:
            self.attributes["script_title"] = self.scripts.script_list[0].title
            for script in self.scripts.script_list:
                available_scripts.append(script.title)
        self.elements.append(new_combo_element(available_scripts,
                                               "Select scripts to run",
                                               self.script_box_activated))

        # 2 - Cell identification number
        self.elements.append(new_text_element("Cell ID",
                                              "Cell identification",
                                              self.set_id))

        # 3 - Path to log file
        self.elements.append(new_text_element(
            "Log file",
            "File to log to, placed in specified logs folder",
            self.set_path
        ))

        # 4 - auto_fill Button
        self.elements.append(new_button_element(
            "AutoFill",
            "Fill log file from cell identification, defaults to [id]A.log",
            self.button_auto_fill
        ))

        # 5 - Mass
        self.elements.append(new_text_element("Mass",
                                              "Mass of Cell",
                                              self.set_mass))

        # 6 - Comment
        self.elements.append(new_text_element("Comment",
                                              "Unparsed Comment",
                                              self.set_comment))

        # 7 - Package
        package_types = ["Pouch",
                         "Coin",
                         "Cylindrical",
                         "Unknown"]
        self.elements.append(new_combo_element(package_types,
                                               "Package Type",
                                               self.package_box_activated))

        # 8 - Cell type
        cell_types = ["Full",
                      "Half",
                      "AnodeHalf",
                      "CathodeHalf",
                      "LithiumLithium",
                      "Symmetric",
                      "Unknown"]
        self.elements.append(new_combo_element(cell_types,
                                               "Cell Type",
                                               self.cell_box_activated))

        # 9 - Requester selection box
        requestor_options = ["Unspecified",
                             "VC",
                             "GE",
                             "LK"]
        self.elements.append(new_combo_element(requestor_options,
                                               "Person starting cycle",
                                               self.requestor_box_activated))

        # 10 - Start button
        self.elements.append(new_button_element(
            "Read Cell",
            "Read Voltage of Connected Cell",
            self.button_read
        ))

        # 11 - Start button
        self.elements.append(new_button_element(
            "Start",
            "Start Cycle",
            self.button_start
        ))

        # 12 - Pause button
        self.elements.append(new_button_element(
            "Pause",
            "Pause Cycle",
            self.button_pause
        ))

        # 13 - Resume button
        self.elements.append(new_button_element(
            "Resume",
            "Resume Cycle",
            self.button_resume
        ))

        # 14 - Stop button
        self.elements.append(new_button_element(
            "Stop",
            "Stop Cycle",
            self.button_stop
        ))

        # 15 Cell Status
        self.elements.append(QLabel())
        self.elements[-1].setStatusTip("Cell status")

        QMetaObject.connectSlotsByName(self)

    # Begin Button Press Definitions
    def button_auto_fill(self):
        self.threadpool.start(workers.AutoFill(self))
        logging.debug("AutoFill Pressed")

    def button_read(self):
        self.threadpool.start(workers.Read(self))
        logging.debug("Read Pressed")

    def button_start(self):
        self.threadpool.start(workers.Control(self, "start", self.scripts))
        logging.debug("Start Pressed")

    def button_pause(self):
        self.threadpool.start(workers.Control(self, "pause", self.scripts))
        logging.debug("Pause Pressed")

    def button_resume(self):
        self.threadpool.start(workers.Control(self, "resume", self.scripts))
        logging.debug("Resume Pressed")

    def button_stop(self):
        self.threadpool.start(workers.Control(self, "stop", self.scripts))
        logging.debug("Stop Pressed")

    # Begin Attribute Assignment Definitions

    def script_box_activated(self, text):
        """Sets object's script to one selected in dropdown"""
        self.attributes["script_title"] = text

    def package_box_activated(self, text):
        """Sets object's package type to one selected in dropdown"""
        self.attributes["package"] = text

    def cell_box_activated(self, text):
        """Sets object's cell type to one selected in dropdown"""
        self.attributes["type"] = text

    def requestor_box_activated(self, text):
        """Sets object's requestor to one selected in dropdown"""
        self.attributes["requestor"] = text

    def set_id(self, text):
        """Set object identification from text box"""
        self.attributes["id"] = text

    def set_comment(self, text):
        """Set object comment from text box"""
        self.attributes["comment"] = text

    def set_path(self, text):
        """Set object path from text box"""
        self.attributes["path"] = text

    def set_mass(self, text):
        """Set object mass from text box"""
        self.attributes["mass"] = text

    def set_protocol(self, text):
        """Set object protocol from text box"""
        self.attributes["protocol_name"] = text

    def paintEvent(self, event):
        style_option = QStyleOption()
        style_option.initFrom(self)
        painter = QPainter(self)
        style = self.style()
        style.drawPrimitive(QStyle.PE_Widget, style_option, painter, self)



"""Methods and object to handle scripts"""

from os import listdir

from PySide2.QtWidgets import QListWidgetItem


class Scripts(object):
    def __init__(self):
        self.script_list = []

    def load_default_scripts(self, path):
        """Load scripts from scripts folder"""
        files = listdir(path)
        if files is not None:
            for file in files:
                self.script_list.append(Script(file, path))

    def get_script_by_title(self, title):
        """Returns script object with given title"""
        for script in self.script_list:
            if script.title == title:
                return script
        return None


class Script(QListWidgetItem):
    """Object to store and manipulate scripts"""
    def __init__(self, title, path):
        super().__init__()
        self.title = title
        self.path = path
        self.content = open(self.path + "/" + self.title, "r").read()
        self.setText(self.title)

    def save(self):
        """Saves script to file"""
        with open(self.path + "/" + self.title, "w") as file:
            file.write(self.content)
        self.update_status()

    def update_status(self):
        """Updates title with '*' if script has been edited"""
        if self.content == open(self.path + "/" + self.title, "r").read():
            self.setText(self.title)
        else:
            self.setText("* " + self.title)



"""Classes to handle interfacing with Keithleys and their channels"""

import time
import visa


def source_from_gpib(gpib_address, channel):
    """Opens GPIB resource and returns as Source object"""
    resource_manager = visa.ResourceManager()
    source_meter = resource_manager.open_resource(
        "GPIB0::{}::INSTR".format(gpib_address)
    )
    return Source(source_meter, channel)


class Keithley2602(object):
    """Represents a single keithley Interface"""

    # TODO something in the startup script screws up CC discahrge.
    # It does constant V discharge instead
    script_startup = open("resources/startup.script").read()

    current_ranges = [100 * 1e-9, 1e-6, 10e-6,
                      100e-6, 1e-3, 0.01,
                      0.1, 1.0, 3.0]

    def __init__(self, gpib_addr, load_scripts=True):
        resource_manager = visa.ResourceManager()
        self.gpib_addr = gpib_addr
        self.source_meter = resource_manager.open_resource(
            "GPIB0::{}::INSTR".format(gpib_addr)
        )
        # TODO do not reset? Do something else, clear buffers I think
        self.source_meter.write("reset()")
        if load_scripts:
            self.source_meter.write(self.script_startup)

    def get_source(self, kch, channel=None):
        """Get source object of Keithley"""
        return Source(self.source_meter, kch, channel=channel)


class Source(object):
    """Represents an individual source"""
    current_ranges = [100 * 1e-9, 1e-6, 10e-6,
                      100e-6, 1e-3, 0.01,
                      0.1, 1.0, 3.0]

    def __init__(self, source_meter, kch, channel=None):
        """

        Parameters
        ----------
        source_meter: sourcemeter
            Obtained from an open_resource visa call
        kch: str
            Keithley channel ("a" or "b").
            Stored lower case internally and accessible in the kch attribute
        channel: int or str
            Channel that the user sees. Can be integer or string,
            however will be used to sort and display channels
            Defaults to None

        Source is typically gnerated from the
        Keithley object's get_source function
        """
        self.source_meter = source_meter
        # This is the Keithley channel on dual channel Keithleys
        # It can be "a" or "b"
        kch = kch.lower()
        if kch == "a":
            self.snum = 1
            self.identification = "smua"
            self.kch = "a"
        elif kch == "b":
            self.snum = 2
            self.identification = "smub"
            self.kch = "b"
        else:
            raise ValueError("ch argument must be 'a' or 'b'")
        self.chd = {"a": 1, "b": 2}

        # this is the channel that the user sees, it should be a string
        #
        self.channel = str(channel)
        self.data = []  # will hold, timestamp, current, voltage
        self.report = []
        self.data_max_len = 500

        # Show the channel number on the screen
        if self.channel:
            self.set_text(text1="Channel {}".format(self.channel))

    def off(self):
        self.source_meter.write(
            "smu{ch}.source.output = smu{ch}.OUTPUT_OFF".format(ch=self.kch)
        )

    def get_range(self, current):
        return min([c for c in self.current_ranges if c > abs(current)])

    def set_current(self, current, v_limit):

        script = """display.screen = display.SMUA_SMUB
display.smu{ch}.measure.func = display.MEASURE_DCVOLTS
smu{ch}.source.func = smu{ch}.OUTPUT_DCAMPS
smu{ch}.source.autorangei = smu{ch}.AUTORANGE_ON
smu{ch}.source.leveli = {current}
smu{ch}.source.limitv = {v_limit}
smu{ch}.source.output = smu{ch}.OUTPUT_ON""".format(ch=self.kch,
                                                    current=current,
                                                    v_limit=v_limit)
        self._run_script(script, "setcurrent")

    def rest(self, v_limit=5.0):
        self.set_current(0.0, v_limit)

    def _run_script(self, script, scriptname):
        instruction = "loadscript {}()\n{}\nendscript\n{}()".format(
            scriptname, script, scriptname
        )
        return self.source_meter.write(instruction)

    def pause(self):
        self.write(
            "smu{ch}.source.output = smu{ch}.OUTPUT_OFF".format(ch=self.kch)
        )

    def set_text(self, text1: str = "", text2: str = ""):
        """
        Unfortunately the Keithley does not treat the two channels
        independently for display purposes. So setting the text for one
        channels removes all the info for the other channel rendering this
        functionality nearly useless.

        Parameters
        ----------
        text1: str
            top line of text, 10 max chars
        text2: str
            bottom line of text, 16 max chars

        Returns
        -------

        """
        ''
        letters = "".join(map(chr, range(65, 91)))
        letters += letters.lower()
        numbers = "".join(map(chr, range(48, 58)))
        others = " :,-_"
        acceptable_chr = letters + numbers + others

        if text1:
            text1 = "".join([c for c in text1 if c in acceptable_chr])[:10]
            startpos = {"a": 1, "b": 11}[self.kch]
            self.write("display.setcursor(1,{})".format(startpos))
            self.write('display.settext("{}")'.format(text1.ljust(10)))

        if text2:
            text2 = "".join([c for c in text2 if c in acceptable_chr])[:16]
            startpos = {"a": 1, "b": 17}[self.kch]
            self.write("display.setcursor(2,{})".format(startpos))
            self.write('display.settext("{}")'.format(text2.ljust(16)))

    def set_voltage(self, voltage, i_limit):
        script = \
            """display.screen = display.SMUA_SMUB
            display.smu{ch}.measure.func = display.MEASURE_DCAMPS
            smu{ch}.source.func = smu{ch}.OUTPUT_DCVOLTS
            smu{ch}.source.autorangei = smu{ch}.AUTORANGE_ON
            smu{ch}.source.levelv = {voltage}
            smu{ch}.source.limiti = {current}
            smu{ch}.source.output = smu{ch}.OUTPUT_ON
            """.format(ch=self.kch, voltage=voltage, current=i_limit)
        self._run_script(script, "setvoltage")

    def read_iv(self):
        self.source_meter.write(
            "current, voltage = smu{}.measure.iv()".format(self.kch)
        )
        current = float(self.source_meter.ask("print(current)"))
        voltage = float(self.source_meter.ask("print(voltage)"))

        # The Keithley will report totally out of range numbers like 9.91e+37
        # if asked to e.g. charge to 3.9V when the cell is already at 4.2V
        # It is basically its way of saying the condition cannot be achieved
        # The actual current sent is 0.0 A.
        if abs(current) > 1.0e10:
            current = 0.0
        return current, voltage

    def read_data(self):
        t = time.time()
        self.source_meter.write(
            "current, voltage = smu{}.measure.iv()".format(self.kch)
        )
        current = float(self.source_meter.ask("print(current)"))
        voltage = float(self.source_meter.ask("print(voltage)"))
        self.data.append([t, current, voltage])

        if len(self.data) > self.data_max_len:
            # we pop 1 and not 0
            # because if we pop 0 it will screw up the total time checking
            self.data.pop(1)

    def read_until(self, write_conditions, end_conditions, wait_time=5.):
        count = 0
        ctn = True
        while ctn:
            self.read_data()
            print(self.data[-1], end='\n', flush=True)
            next_time = self.data[-1][0] + wait_time
            for condition in end_conditions:
                if condition.check(self.data):
                    ctn = False
                    break
                else:
                    next_time = min(condition.next_time, next_time)

            if ctn:
                for condition in write_conditions:
                    if condition.check(self.data, self.report):
                        self.report.append(self.data[-1])
                        print('\nwriting {}'.format(self.report[-1]),
                              flush=True)
                        break
            else:
                self.report.append(self.data[-1])
                print('\nwriting')
                break

            actual_wait = next_time - time.time()
            print('waiting: {}'.format(actual_wait), end='\n', flush=True)
            if actual_wait > 0:
                time.sleep(actual_wait)
            count += 1

    def write(self, *args, **kwargs):
        self.source_meter.write(*args, **kwargs)

    def ask(self, *args, **kwargs):
        self.source_meter.ask(*args, **kwargs)



"""Main window for the cyckei client."""

import logging

from PySide2.QtWidgets import QWidget, QMainWindow, QAction, qApp, QTabWidget,\
    QMessageBox
from PySide2.QtGui import QIcon
from PySide2.QtCore import QThreadPool

from cyckei.client.channel_tab import ChannelTab
from cyckei.client.script_tab import ScriptEditor
from cyckei.client.log_tab import LogViewer
from workers import Ping
from scripts import Scripts


def help():
    """Direct to help"""
    msg = QMessageBox()
    msg.setIcon(QMessageBox.Information)
    msg.setText(
        """For help refer to the HELP.md and README.md files located in
        the cyckei install location, or online on our GitLab page.
        \n\ngitlab.com/cyclikal/cyckei"""
    )
    msg.setWindowTitle("Help")
    msg.exec_()


def about():
    """Display basic information about cyckei"""
    msg = QMessageBox()
    msg.setIcon(QMessageBox.Information)
    msg.setText(
        """Cyckei is developed by Gabriel Ewig and Vincent
        Chevrier at Cyclikal, LLC.\n\nUpdates and source code can be found on
        GitLab at gitlab.com/cyclikal/cyckei.\n\nFor information about
        Cyclikal, visit cyclikal.com."""
    )
    msg.setWindowTitle("About")
    msg.exec_()


class MainWindow(QMainWindow):
    """Main Window class which is and sets up itself"""
    # Setup main windows
    def __init__(self, config):
        super().__init__()
        # Set basic window properties
        self.setWindowTitle("Cyckei")
        self.config = config
        # Allows icon to show in Windows Taskbar
        self.setWindowIcon(QIcon("resources/icon.png"))
        self.resize(1000, 562)

        # Setup ThreadPool
        self.threadpool = QThreadPool()
        logging.info("Multithreading set with maximum {} threads".format(
            self.threadpool.maxThreadCount()
        ))

        # Load scripts
        scripts = Scripts()
        scripts.load_default_scripts(config["path"] + "/scripts")

        # Create menu and status bar
        self.menu_bar = self.create_menu()
        self.status_bar = self.statusBar()

        self.tab_widget = QTabWidget(self)
        self.setCentralWidget(self.tab_widget)

        self.tab_widget.addTab(
            ChannelTab(self.config, self.threadpool, scripts),
            "Channels"
        )
        self.tab_widget.addTab(
            ScriptEditor(self.tab_widget.widget(0).channels, scripts),
            "Scripts"
        )
        self.tab_widget.addTab(LogViewer(self.config), "Logs")

        self.setStyleSheet(open("resources/style.css", "r").read())

    def create_menu(self):
        """Setup menu bar"""
        menu_options = []

        menu_options.append(QAction("&About", self))
        menu_options[-1].setStatusTip("About Cyckei")
        menu_options[-1].triggered.connect(about)

        menu_options.append(QAction("&Help", self))
        menu_options[-1].setStatusTip("View Help File")
        menu_options[-1].triggered.connect(help)

        menu_options.append(QAction("&Exit", self))
        menu_options[-1].setStatusTip("Exit Client Application")
        menu_options[-1].triggered.connect(qApp.quit)

        menu_options.append(QAction("&Ping", self))
        menu_options[-1].setStatusTip("Ping Server for Response")
        menu_options[-1].triggered.connect(self.ping_server)

        menu_options.append(QAction("&Save", self))
        menu_options[-1].setStatusTip("Save Batch of IDs and Log Files")
        menu_options[-1].triggered.connect(self.save_batch)

        menu_options.append(QAction("&Load", self))
        menu_options[-1].setStatusTip("Load Batch of IDs and Log Files")
        menu_options[-1].triggered.connect(self.load_batch)

        menu_options.append(QAction("&Fill All", self))
        menu_options[-1].setStatusTip("Auto Fill All Log Files in Batch")
        menu_options[-1].triggered.connect(self.fill_batch)

        menu_options.append(QAction("&Increment", self))
        menu_options[-1].setStatusTip("Increment Batch")
        menu_options[-1].triggered.connect(self.increment_batch)

        menu_bar = self.menuBar()
        menu_bar.setNativeMenuBar(False)

        menu = menu_bar.addMenu("&Menu")
        menu.addAction(menu_options[0])
        menu.addAction(menu_options[1])
        menu.addAction(menu_options[2])

        server = menu_bar.addMenu("&Server")
        server.addAction(menu_options[3])

        batch = menu_bar.addMenu("&Batch")
        batch.addAction(menu_options[4])
        batch.addAction(menu_options[5])
        batch.addAction(menu_options[6])
        batch.addAction(menu_options[7])

        return menu_bar

    def ping_server(self):
        self.threadpool.start(Ping())

    def save_batch(self):
        """Saves id and log information to file"""
        reply = QMessageBox.question(
            QWidget(),
            "Batch",
            "Save batch?\nCurrent saved batch will be deleted.",
            QMessageBox.Yes | QMessageBox.No,
            QMessageBox.Yes
        )
        if reply == QMessageBox.Yes:
            batch = []
            for channel_widget in self.channels:
                channel = [channel_widget.attributes["id"],
                           channel_widget.attributes["path"]]
                batch.append(channel)

            with open(self.config["path"] + "/batch.txt", "a") as file:
                file.truncate(0)
                for channel in batch:
                    for value in channel:
                        file.write(str(value) + ",")
                    file.write("\n")

    def load_batch(self):
        """Loads id and log information from file"""
        reply = QMessageBox.question(
            QWidget(),
            "Batch",
            "Load batch?\nAll current values will be overwritten.",
            QMessageBox.Yes | QMessageBox.No,
            QMessageBox.Yes
        )
        if reply == QMessageBox.Yes:
            with open(self.config["path"] + "/batch.txt", "r") as file:
                for index, line in enumerate(file):
                    channel = self.channels[index]
                    values = line.split(",")
                    if len(values) > 1:
                        channel.elements[2].setText(values[0])
                        channel.elements[3].setText(values[1])

    def fill_batch(self):
        """Executes autofill for each channel"""
        for channel in self.channels:
            channel.auto_fill()

    def increment_batch(self):
        """Increments last letter of batch by char number"""
        for channel in self.channels:
            working_list = list(channel.elements[3].text())
            if working_list:
                try:
                    current_letter = ord(working_list[-5])
                    next_letter = chr(current_letter + 1)
                    working_list[-5] = next_letter

                    channel.elements[3].setText("".join(working_list))
                except Exception as exception:
                    msg = QMessageBox()
                    msg.setText(
                        "Could not increment channel {}.\n".format(
                            channel.channel
                        )
                        + str(exception)
                    )
                    msg.setIcon(QMessageBox.Warning)
                    msg.setWindowTitle("Increment")
                    msg.exec_()



from setuptools import find_packages, setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="cyckei",
    version="0.0.1.dev4",
    author="Gabriel Ewig",
    author_email="gabriel@cyclikal.com",
    description="Keithley Battery Cycler for Python3",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords='keithley battery lithium-ion cycling calorimetry',
    url="https://gitlab.com/cyclikal/cyckei",

    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering",
    ],
    install_requires=[
        "zmq",
        "pyvisa",
        "PySide2",
    ],
    entry_points={
        "console_scripts": [
            "cyckei = cyckei.__main__:main"
        ],
    },
)



import json
import logging

from PySide2.QtCore import QRunnable, Slot
from PySide2.QtWidgets import QMessageBox
from os.path import exists
from os import makedirs, remove
from datetime import date
from time import sleep

from socket import Socket


def not_none(value):
    """Sets a None value to "None" string"""
    return "None" if value is None else str(value)


def send(json):
    """Sends json to server and updates status with response"""
    # TODO: Load info from config
    socket = Socket("tcp://localhost", 5556)
    resp = socket.send(json)["response"]
    logging.info(resp)
    socket.socket.close()
    return resp


def prepare_json(channel, function, scripts):
    """Sets the channel's json script to current values"""
    protocol = scripts.get_script_by_title(
        channel.attributes["script_title"]).content
    json_packet = json.load(open("resources/defaultJSON.json"))

    json_packet["function"] = function
    json_packet["kwargs"]["channel"] = channel.attributes["channel"]
    json_packet["kwargs"]["meta"]["cellid"] = channel.attributes["id"]
    json_packet["kwargs"]["meta"]["comment"] = channel.attributes["comment"]
    json_packet["kwargs"]["meta"]["package"] = channel.attributes["package"]
    json_packet["kwargs"]["meta"]["cell_type"] = channel.attributes["type"]
    temp_path = (
        channel.attributes["record_folder"]
        + "/"
        + str(date.today())
    )
    if not exists(temp_path):
        makedirs(temp_path)
    json_packet["kwargs"]["meta"]["path"] = (
        temp_path
        + "/"
        + channel.attributes["path"]
    )
    json_packet["kwargs"]["meta"]["mass"] = channel.attributes["mass"]
    json_packet["kwargs"]["meta"]["protocol_name"]\
        = channel.attributes["script_title"]
    json_packet["kwargs"]["meta"]["requester"]\
        = channel.attributes["requestor"]
    json_packet["kwargs"]["meta"]["channel"] = channel.attributes["channel"]
    json_packet["kwargs"]["meta"]["protocol"] = protocol
    json_packet["kwargs"]["protocol"] = protocol

    return json_packet


class Ping(QRunnable):
    def __init__(self):
        super(Ping, self).__init__()

    @Slot()
    def run(self):
        # TODO: Load info from config
        Socket("tcp://localhost", 5556).ping_server()


class UpdateStatus(QRunnable):
    """Update status shown below controls by contacting server"""
    def __init__(self, channel):
        super(UpdateStatus, self).__init__()
        self.channel = channel

    @Slot()
    def run(self):
        # TODO: Utilize configuration
        socket = Socket("tcp://localhost", 5556)
        while True:
            info_channel = socket.info_channel(
                self.channel.attributes["channel"])["response"]
            channel_status = socket.channel_status(
                self.channel.attributes["channel"])["response"]
            try:
                status = (channel_status
                          + " - " + not_none(info_channel["state"])
                          + " | C: " + not_none(info_channel["current"])
                          + ", V: " + not_none(info_channel["voltage"]))
            except TypeError:
                status = info_channel
            logging.debug("Updating channel {} with satus {}".format(
                self.channel.attributes["channel"], status))
            self.channel.elements[15].setText(status)
            sleep(2)


class AutoFill(QRunnable):
    """Fill log text with value derived from cell identification"""
    def __init__(self, channel):
        super(AutoFill, self).__init__()
        self.channel = channel

    @Slot()
    def run(self):
        if self.channel.elements[2].text():
            self.channel.elements[3].setText(
                "{}A.pyb".format(self.elements[2].text()))


class Read(QRunnable):
    """Tell channel to Rest() long enough to get voltage reading on cell"""
    def __init__(self, channel):
        super(Read, self).__init__()
        self.channel = channel

    @Slot()
    def run(self):
        socket = Socket("tcp://localhost", 5556)
        package = json.load(open("resources/defaultJSON.json"))
        package["function"] = "start"
        package["kwargs"]["channel"] = self.channel.attributes["channel"]
        package["kwargs"]["meta"]["path"] = (
            self.channel.attributes["record_folder"]
            + "/{}.temp".format(self.channel.attributes["channel"])
        )
        package["kwargs"]["protocol"] = """Rest()"""
        socket.send(package)

        info_channel = socket.info_channel(
            self.channel.attributes["channel"])["response"]
        try:
            status = ("Voltage of cell: "
                      + not_none(info_channel["voltage"]))
        except Exception:
            status = "Could not read cell voltage."

        package["function"] = "stop"
        socket.send(package)

        self.channel.elements[-1].setText(status)

        socket.socket.close()


class Control(QRunnable):
    """Update json and send "start" function to server"""
    def __init__(self, channel, command, scripts):
        super(Control, self).__init__()
        self.channel = channel
        self.command = command
        self.scripts = scripts

    @Slot()
    def run(self):
        if self.command == "start":
            script_ok = Check(self.scripts.get_script_by_title(
                    self.channel.attributes["script_title"]).content).run()
            if script_ok is False:
                return

        response = send(prepare_json(self.channel, self.command, self.scripts))
        self.channel.elements[-1].setText(response)


class Check(QRunnable):
    def __init__(self, protocol):
        super(Check, self).__init__()
        self.protocol = protocol

    @Slot()
    def run(self):
        """Initiates checking tests"""
        passed, msg = self.legal_test(self.protocol)
        if not passed:
            return self.end_false(msg)
        passed, msg = self.run_test(self.protocol)
        if not passed:
            return self.end_false(msg)
        return True

    def legal_test(self, protocol):
        """Checks if script only contains valid commands"""
        conditions = ["#",
                      "for",
                      "AdvanceCycle()",
                      "CCCharge(",
                      "CCDischarge(",
                      "CVCharge(",
                      "CVDischarge(",
                      "Rest(",
                      "Sleep("]

        for line in protocol.splitlines():
            line = line.replace(" ", "")
            line = line.replace("\t", "")

            valid = False
            for condition in conditions:
                if line.startswith(condition) or not line:
                    valid = True
                    break
            if not valid:
                return False, "Illegal command: \"" + line + "\"."

        return True, "Passed"

    def run_test(self, protocol):
        """Checks if server can load script successfully"""
        packet = self.prepare_json(protocol)
        response = send(packet)
        if response == "Passed":
            return True, "Passed"
        return (False,
                "Server failed to run script. Error: \"{}\".".format(response))

    def end_false(self, reason):
        """Show message box with error statement and return false"""
        msg = QMessageBox()
        msg.setIcon(QMessageBox.Warning)
        msg.setText("Failed!")
        msg.setInformativeText("Script did not pass the check.")
        msg.setWindowTitle("Check Failed")
        msg.setDetailedText(reason)
        msg.exec_()
        return False

    def prepare_json(self, protocol):
        """create json to send to server"""
        json_packet = json.load(open("resources/defaultJSON.json"))

        json_packet["function"] = "test"
        json_packet["kwargs"]["protocol"] = protocol

        return json_packet
